Package: fable.bayesRecon
Date: 2026-05-05
Title: Bayesian Reconciliation in the 'fable' Framework
Version: 0.1.0
Authors@R: 
    c(person(given = "Dario",
             family = "Azzimonti",
             role = c("aut","cre","cph"),
             email = "dario.azzimonti@gmail.com",
             comment = c(ORCID = "0000-0001-5080-3061")),
      person(given = "Stefano",
             family = "Damato",
             role = c("aut"),
             email = "stefano.damato@idsia.ch",
             comment = c(ORCID = "0009-0001-0225-3736")),
      person(given = "Lorenzo",
             family = "Zambon",
             role = c("aut"),
             email = "lorenzo.zambon@idsia.ch",
             comment = c(ORCID = "0000-0002-8939-993X")),
      person(given = "Chiara",
             family = "Carrara",
             role = c("aut"),
             email = "chiara.carrara03@universitadipavia.it",
             comment = c(ORCID = "0009-0002-7349-6249")),       
      person(given = "Giorgio",
             family = "Corani",
             role = c("aut"),
             email = "giorgio.corani@idsia.ch",
             comment = c(ORCID = "0000-0002-1541-8384")))
Description: Implements the 'bayesRecon' probabilistic reconciliation methods
    within the 'fable' framework for hierarchical time series forecasting.
    Bayesian reconciliation (bayesRecon) methods are accessed via the 'reconcile' verb, following
    'fable' conventions. For methodological background, see Corani et al. (2021)
    <doi:10.1007/978-3-030-67664-3_13>, Zambon et al. (2024a)
    <doi:10.1007/s11222-023-10343-y>, Zambon et al. (2024b)
    <https://proceedings.mlr.press/v244/zambon24a.html>, and Carrara et al.
    (2025) <doi:10.48550/arXiv.2506.19554>.
License: LGPL (>= 3)
URL: https://github.com/dazzimonti/fable.bayesRecon
BugReports: https://github.com/dazzimonti/fable.bayesRecon/issues
Encoding: UTF-8
Imports: stats, bayesRecon, fabletools (>= 0.7.0), purrr, vctrs,
        distributional, dplyr, tsibble, rlang
Suggests: testthat (>= 3.0.0), cli, fable, tsibbledata, tibble,
        ggplot2, scales, knitr, rmarkdown
VignetteBuilder: knitr
Config/testthat/edition: 3
Depends: R (>= 4.1)
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-05-21 13:56:12 UTC; dario.azzimonti
Author: Dario Azzimonti [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0001-5080-3061>),
  Stefano Damato [aut] (ORCID: <https://orcid.org/0009-0001-0225-3736>),
  Lorenzo Zambon [aut] (ORCID: <https://orcid.org/0000-0002-8939-993X>),
  Chiara Carrara [aut] (ORCID: <https://orcid.org/0009-0002-7349-6249>),
  Giorgio Corani [aut] (ORCID: <https://orcid.org/0000-0002-1541-8384>)
Maintainer: Dario Azzimonti <dario.azzimonti@gmail.com>
Repository: CRAN
Date/Publication: 2026-05-28 12:20:06 UTC
Built: R 4.6.0; ; 2026-05-28 21:37:05 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(forecast,lst_bayesRecon_BUIS)
S3method(forecast,lst_bayesRecon_MixCond)
S3method(forecast,lst_bayesRecon_TDcond)
S3method(forecast,lst_bayesRecon_t)
export(bayesRecon_BUIS)
export(bayesRecon_MixCond)
export(bayesRecon_TDcond)
export(bayesRecon_t)
importFrom(bayesRecon,.check_hierarchical)
importFrom(bayesRecon,.compute_naive_cov)
importFrom(bayesRecon,.core_reconc_BUIS)
importFrom(bayesRecon,.core_reconc_MixCond)
importFrom(bayesRecon,.core_reconc_TDcond)
importFrom(bayesRecon,.core_reconc_t)
importFrom(bayesRecon,.get_HG)
importFrom(bayesRecon,multi_log_score_optimization)
importFrom(bayesRecon,schaferStrimmer_cov)
importFrom(distributional,cdf)
importFrom(distributional,dist_inflated)
importFrom(distributional,dist_sample)
importFrom(distributional,dist_student_t)
importFrom(distributional,dist_truncated)
importFrom(distributional,generate)
importFrom(distributional,support)
importFrom(dplyr,full_join)
importFrom(fabletools,distribution_var)
importFrom(fabletools,forecast)
importFrom(purrr,exec)
importFrom(purrr,list_c)
importFrom(purrr,map)
importFrom(purrr,map2)
importFrom(purrr,map_dbl)
importFrom(purrr,map_int)
importFrom(purrr,pmap)
importFrom(purrr,reduce)
importFrom(rlang,abort)
importFrom(rlang,enquo)
importFrom(stats,complete.cases)
importFrom(stats,cov2cor)
importFrom(stats,density)
importFrom(stats,family)
importFrom(stats,frequency)
importFrom(stats,quantile)
importFrom(stats,residuals)
importFrom(tsibble,index_var)
importFrom(vctrs,field)
importFrom(vctrs,vec_c)
importFrom(vctrs,vec_slice)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# fable.bayesRecon: BAyesian reCONciliation in the fable framework <a href="https://idsia.github.io/bayesRecon/"><img src="man/figures/logo.png" align="right" height="150" alt="bayesRecon website" /></a>

<!-- badges: start -->

[![R-CMD-check](https://github.com/dazzimonti/fable.bayesRecon/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/dazzimonti/fable.bayesRecon/actions/workflows/R-CMD-check.yaml)
[![CRAN
status](https://www.r-pkg.org/badges/version/fable.bayesRecon)](https://CRAN.R-project.org/package=fable.bayesRecon)
[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![Coverage
Status](https://coveralls.io/repos/github/dazzimonti/fable.bayesRecon/badge.svg?branch=main)](https://coveralls.io/github/dazzimonti/fable.bayesRecon?branch=main)
[![License: LGPL (\>=
3)](https://img.shields.io/badge/license-LGPL%20(%3E%3D%203)-yellow.svg)](https://www.gnu.org/licences/lgpl-3.0)
[![R-CMD-check](https://github.com/dazzimonti/fable.bayesRecon/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/dazzimonti/fable.bayesRecon/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

The package `fable.bayesRecon` integrates the probabilistic
reconciliation methods from
[`bayesRecon`](https://github.com/IDSIA/bayesRecon) into the
[`fable`](https://fable.tidyverts.org/) /
[`fabletools`](https://fabletools.tidyverts.org/) framework.
Reconciliation is specified via the `reconcile()` verb and produced when
`forecast()` is called, following the same tidy workflow used by
`fable`.

The reconciliation functions are:

- `bayesRecon_t`: reconciliation via conditioning with uncertain
  covariance matrix; the reconciled forecasts are multivariate
  Student-t; this is done analytically.
- `bayesRecon_BUIS`: reconciliation via conditioning of any
  probabilistic forecast via importance sampling; this is the
  recommended option for non-Gaussian base forecasts;
- `bayesRecon_MixCond`: reconciliation via conditioning of mixed
  hierarchies, where the upper forecasts are multivariate Gaussian and
  the bottom forecasts are discrete distributions;
- `bayesRecon_TDcond`: reconciliation via top-down conditioning of mixed
  hierarchies, where the upper forecasts are multivariate Gaussian and
  the bottom forecasts are discrete distributions;

## News

:boom: \[2026-05-05\] fable.bayesRecon v0.1.0: first CRAN release.

## Installation

You can install the **development** version from
[GitHub](https://github.com/dazzimonti/fable.bayesRecon):

``` r
# install.packages("devtools")
devtools::install_github("dazzimonti/fable.bayesRecon", build_vignettes = TRUE, dependencies = TRUE)
```

## Usage

The package follows the standard `fable` workflow:

1.  Prepare data as a `tsibble` and define the hierarchy with
    `aggregate_key()`.
2.  Fit base forecasting models with `model()`.
3.  Specify the reconciliation strategy inside `reconcile()`.
4.  Produce reconciled probabilistic forecasts with `forecast()`.

We provide in [this vignette](vignettes/fable.bayesRecon.Rmd) a simple
usage example; refer to the package documentation for more details on
the reconciliation methods and their parameters. See the book Hyndman
and Athanasopoulos (2021) for a general introduction to forecasting with
`fable` and `fabletools`.

## References

Carrara, C., Corani, G., Azzimonti, D., Zambon, L. (2025). *Modeling the
uncertainty on the covariance matrix for probabilistic forecast
reconciliation*. arXiv preprint arXiv:2506.19554. [Available
here](https://arxiv.org/abs/2506.19554)

Hyndman, R.J., & Athanasopoulos, G. (2021). *Forecasting: principles and
practice*. 3rd edition, OTexts: Melbourne, Australia.
[OTexts.com/fpp3](https://OTexts.com/fpp3/). Accessed on 05/05/2026.

Zambon, L., Azzimonti, D. & Corani, G. (2024). *Efficient probabilistic
reconciliation of forecasts for real-valued and count time series*.
Statistics and Computing 34 (1), 21.
[DOI](https://doi.org/10.1007/s11222-023-10343-y)

Zambon, L., Azzimonti, D., Rubattu, N., Corani, G. (2024).
*Probabilistic reconciliation of mixed-type hierarchical time series*.
Proceedings of the Fortieth Conference on Uncertainty in Artificial
Intelligence, PMLR 244:4078-4095. [Available
here](https://proceedings.mlr.press/v244/zambon24a.html)

## Contributors

<!-- prettier-ignore-start -->

<!-- markdownlint-disable -->

<table>

<tbody>

<tr>

<td align="center" valign="top" width="14.28%">

<a href="https://dazzimonti.github.io/dazzimonti/">
<img src="https://github.com/dazzimonti.png" width="100px;" alt="Dario Azzimonti" style="border-radius:50%;border:1px solid #646464;"/><br />
<sub><b>Dario Azzimonti</b></sub></a><br />
<sub>(Maintainer)</sub><br />
<a href="mailto:dario.azzimonti@gmail.com?subject=[fable.bayesRecon package]">Email</a>
</td>

<td align="center" valign="top" width="14.28%">

<a href="#">
<img src="https://github.com/StefanoDamato.png" width="100px;" alt="Stefano Damato" style="border-radius:50%;border:1px solid #646464;"/><br />
<sub><b>Stefano Damato</b></sub></a><br /> <sub> </sub><br />
<a href="mailto:stefano.damato@idsia.ch?subject=[fable.bayesRecon package]">Email</a>
</td>

<td align="center" valign="top" width="14.28%">

<a href="#">
<img src="https://github.com/LorenzoZambon.png" width="100px;" alt="Lorenzo Zambon" style="border-radius:50%;border:1px solid #646464;"/><br />
<sub><b>Lorenzo Zambon</b></sub></a><br /> <sub> </sub><br />
<a href="mailto:lorenzo.zambon@idsia.ch?subject=[fable.bayesRecon package]">Email</a>
</td>

<td align="center" valign="top" width="14.28%">

<a href="#">
<img src="https://github.com/CarraraChiara.png" width="100px;" alt="Chiara Carrara" style="border-radius:50%;border:1px solid #646464;"/><br />
<sub><b>Chiara Carrara</b></sub></a><br /> <sub> </sub><br />
<a href="mailto:chiara.carrara03@universitadipavia.it?subject=[fable.bayesRecon package]">Email</a>
</td>

<td align="center" valign="top" width="14.28%">

<a href="https://sites.google.com/site/awerbhjkl678214/home">
<img src="https://github.com/gcorani.png" width="100px;" alt="Giorgio Corani" style="border-radius:50%;border:1px solid #646464;"/><br />
<sub><b>Giorgio Corani</b></sub></a><br /> <sub> </sub><br />
<a href="mailto:giorgio.corani@idsia.ch?subject=[fable.bayesRecon package]">Email</a>
</td>

</tr>

</tbody>

</table>

<!-- markdownlint-restore -->

<!-- prettier-ignore-end -->

## Getting help

If you encounter a bug, please file a minimal reproducible example on
[GitHub](https://github.com/dazzimonti/fable.bayesRecon/issues).

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(echo = TRUE, warning = FALSE, message = FALSE)
options(tibble.width = Inf, knitr.kable.NA = "")

# Helper to render fabletools tables as HTML-safe kable
kable <- function(x, ...) {
  x |>
    dplyr::mutate(dplyr::across(
      where(~ inherits(.x, "agg_vec")),
      ~ gsub("<", "&lt;", gsub(">", "&gt;", as.character(.x)))
    )) |>
    knitr::kable(...)
}

## ----libraries----------------------------------------------------------------
library(bayesRecon)
library(dplyr)
library(tsibble)
library(fable)
library(fabletools)
library(fable.bayesRecon)
library(ggplot2)
library(scales)

## ----data---------------------------------------------------------------------
data <- swiss_tourism$ts[, -1, drop = FALSE] |>             # Remove 1st column
  as_tsibble() |>                                           # Convert in tsibble
  rename(Month = index, Canton = key, Tourists = value) |>  # Rename
  aggregate_key(Canton, Tourists = sum(Tourists))           # Aggregate all cantons

data |>  head(5) |> kable()

## ----fit----------------------------------------------------------------------
fit <- data |>
  filter(Month < yearmonth("2024 Feb")) |>                  # Remove test set
  model(base = ETS(Tourists ~ trend("A") + season("A"))) |> # fit ETS model
  reconcile(t = bayesRecon_t(base),                         # Reconcile with t-Rec
            mint = min_trace(base))                         # Reconcile with MinT

fit |> head(5) |> kable()

## ----forecast-----------------------------------------------------------------
fc <- fit |>
  forecast(h = "1 year")

# Showing only the one step ahead forecast
fc |> filter(Month == yearmonth("Feb 2024")) |> head(5) |> kable()

## ----accuracy-----------------------------------------------------------------
results <- fc |>
  accuracy(data, measures = list(RMSSE = RMSSE, MASE = MASE, CRPS = CRPS)) |>
  group_by(.model) |>
  summarise(RMSSE = mean(RMSSE), MASE  = mean(MASE), CRPS  = mean(CRPS))

results |> kable(digits = 3)

## ----plot-aggregated, fig.width=7, fig.height=4-------------------------------
# Aggregated series
fc |>
  filter(is_aggregated(Canton)) |>
  autoplot(data, level=95) +
  scale_y_continuous(labels = label_number(suffix = "K", scale = 1e-3))+
  labs(title = "Aggregated Swiss Tourism", y = "Tourists", x = NULL) +
  theme_minimal()

## ----plot-aggregated-zoom, fig.width=7, fig.height=4--------------------------
# Zoom on the forecasts for the aggregated series
fc |>
  filter(is_aggregated(Canton)) |>
  autoplot(data |> filter(Month >= yearmonth("2022 Jan")), level = 95) +
  scale_y_continuous(labels = label_number(suffix = "K", scale = 1e-3))+
  labs(title = "Aggregated Swiss Tourism (Jan 2022 - Jan 2025)", y = "Tourists", x = NULL) +
  theme_minimal()

## ----plot-canton1, fig.width=7, fig.height=4----------------------------------
# Canton Ticino
fc |>
  filter(Canton == "Ticino") |>
  autoplot(data |> filter(Month >= yearmonth("2022 Jan")),
           level=95) +
  scale_y_continuous(labels = label_number(suffix = "K", scale = 1e-3))+
  labs(title = paste("Canton: Ticino"), y = "Tourists", x = NULL) +
  theme_minimal()

## ----plot-canton2, fig.width=7, fig.height=4----------------------------------
# Canton Luzern
fc |>
  filter(Canton == "Luzern") |>
  autoplot(data |> filter(Month >= yearmonth("2022 Jan"),
                                 Canton == "Luzern"),
           level = 80) +
  scale_y_continuous(labels = label_number(suffix = "K", scale = 1e-3))+
  labs(title = paste("Canton: Luzern"), y = "Tourists", x = NULL) +
  theme_minimal()

#'  t-Rec: reconciliation via conditioning with uncertain covariance via multivariate t-distribution
#'
#' @description
#' 
#' Specifies t-Rec reconciliation for use within `reconcile()`. 
#' 
#' Reconciles base forecasts by conditioning on the hierarchical constraints. 
#' The base forecasts are assumed to be jointly Gaussian, 
#' conditionally on the covariance matrix of the forecast errors. 
#' A Bayesian approach is adopted to account for the uncertainty of the covariance matrix. 
#' An Inverse-Wishart prior is specified on the covariance matrix, 
#' leading to a multivariate t-distribution for the base forecasts. 
#' The reconciliation via conditioning is in closed-form, 
#' yielding a multivariate t reconciled distribution.
#' 
#' Reconciliation is performed when `forecast()` is called on the resulting model.
#' Marginal reconciled forecasts follow a Student-t distribution.
#' 
#'
#' @param models A list of fitted models to reconcile.
#' @param ... Additional arguments passed to other methods, including:  
#' - `prior`: Optional list with entries `nu` and `Psi` specifying the parameters 
#' of the Inverse-Wishart prior distribution for the covariance matrix. 
#' If not provided, the prior is estimated from the data.
#' - `freq`: Optional frequency of the time series, used for estimating the naive covariance matrix
#' via seasonal naive residuals. If not provided, the frequency is inferred from the data.
#' - `criterion`: Criterion for estimating the naive covariance matrix (default: "RSS").
#' - `l_shr`: Optional shrinkage parameter (between 0 and 1) for the covariance matrix of the residuals. 
#'
#' @return A model specification of class \code{"bayesRecon_t"}
#'   (inheriting from \code{"mdl_lst"}) to be passed to
#'   \code{\link[fabletools]{reconcile}}. The reconciliation is
#'   performed when \code{\link[fabletools]{forecast}} is called on
#'   the resulting mable (a model table). The forecast output is a \code{fable} object
#'   containing reconciled probabilistic forecasts represented as
#'   sample distributions (\code{\link[distributional]{dist_student_t}}).
#'   
#' @examples 
#' library(fable)
#' library(fabletools)
#' library(tsibble)
#' library(fable.bayesRecon)
#' 
#' # Small hierarchical example using tourism data
#' tourism_small <- tsibble::tourism |>
#' dplyr::filter(Region == "Melbourne") |> 
#' fabletools::aggregate_key(Purpose, Trips = sum(Trips))
#' 
#' fit <- tourism_small |> 
#' model(base = ETS(Trips ~ trend("A") + season("A"))) |>
#' reconcile(recon = bayesRecon_t(base))
#' 
#' fc <- forecast(fit, h = "2 years")
#' fc
#'
#'
#' @references
#' Carrara, C., Corani, G., Azzimonti, D., & Zambon, L. (2025). Modeling the uncertainty on the covariance
#' matrix for probabilistic forecast reconciliation. arXiv preprint arXiv:2506.19554.
#' \url{https://arxiv.org/abs/2506.19554}
#' 
#' @seealso [fabletools::reconcile()], [fabletools::aggregate_key()],
#'  [fabletools::min_trace()], [bayesRecon::reconc_t()]
#'
#' @export
bayesRecon_t <- function(models, ...) {
  structure(models, class = c("lst_bayesRecon_t", "mdl_lst", "list"),
            ...)
}

#' forecast.lst_bayesRecon_t
#'
#' Produces probabilistic forecasts reconciled via Student-t reconciliation.
#' The method reconciles base forecasts from upper and bottom level series
#' using a multivariate Student-t model with flexible prior/posterior specification.
#'
#' @importFrom fabletools forecast distribution_var
#' @importFrom stats family frequency cov2cor
#' @importFrom purrr map map_dbl map_int list_c reduce exec
#' @importFrom bayesRecon .core_reconc_t multi_log_score_optimization .compute_naive_cov
#' @importFrom dplyr full_join
#' @importFrom tsibble index_var
#' @importFrom distributional dist_student_t
#'
#' @method forecast lst_bayesRecon_t
#'
#' @param object An object of class `lst_bayesRecon_t` containing fitted  models.
#' @param key_data A keyed data frame from `fabletools`.
#' @param point_forecast A list of point forecast functions (default: `list(.mean = mean)`).
#' @param new_data Optional new data for forecasting (not currently used).
#' @param ... Additional arguments passed to other methods.
#'
#' @return A fable object with t-reconciled distributions and point forecasts.
#' 
#' @keywords internal
#' @export
forecast.lst_bayesRecon_t <- function(
    object,
    key_data,
    point_forecast = list(.mean = mean),
    new_data = NULL,
    ...
) {
  # Take models from fabletools, and prepare for BUIS
  # Produce the structural matrix from the key_data structure
  S <- fabletools::coherent_smat(key_data)
  
  # applies the next method ("mdl_lst", in class definition above) to extract the fitted models.
  fc <- NextMethod()
  
  # Series of lapply to extract the parameters of the distribution
  fc_dist <- lapply(fc, function(x) x[[distribution_var(x)]])

  ## START OUR REWRITE OF bayesRecon_t WITH distributional
  hier <- get_hier(S, fc_dist)
  A <- hier$A
  base_forecast_h <- hier$base_forecast_h
  upr_ts <- hier$upr_ts
  btm_ts <- hier$btm_ts
  btm_idx <- hier$btm_idx
  n_upr <- hier$n_upr
  n_tot <- hier$n_tot

  # Check that base forecasts are Normal
  if (!all(map(base_forecast_h, family) |> list_c() == "normal")) {
    stop("t-reconciliation works under the assumption of Normal forecasts")
  }
  
  # Extract additional parameters specified as attributes 
  prior <- attr(object, "prior")
  freq <- attr(object, "freq")
  criterion <- attr(object, "criterion")
  l_shr <- attr(object, "l_shr")
  if (is.null(criterion)) criterion <- "RSS"
  if (is.null(l_shr)) l_shr <- 1e-04
  
  # Compute the covariance matrix of the residuals
  res <- get_residuals(object, upr_ts, btm_ts, btm_idx, n_upr)
  covm_res <- crossprod(res) / nrow(res) 
  R1 <- cov2cor(covm_res)
    
  # Try to get directly the prior from the argument
  if (!is.null(prior)) {
    if (is.list(prior)) {
      nu_prior = prior$nu
      Psi_prior = prior$Psi
      if (is.null(nu_prior) | is.null(Psi_prior)) {
        stop("Input error: prior must be a list with entries nu and Psi")
      }
    } else {
      stop("Input error: prior must be a list with entries nu and Psi")
    }
  } else {
    
    # Compute the prior from observed data; obtain them in matrix format
    obs = map(object[c(upr_ts, btm_ts[btm_idx])], ~.$data)
    if(length(unique(map_dbl(obs, nrow))) > 1){
      obs <- unname(as.matrix(reduce(obs, full_join, by = index_var(obs[[1]]))[,-1]))
    } else {
      obs <- matrix(exec(c, !!!map(obs, `[[`, 2)), ncol = length(object))
    }
      
    # Identify the frequency and compute the residuals of the naive forecasts
    if (is.null(freq)){
      freq <- map_int(object[c(upr_ts, btm_ts[btm_idx])], ~ frequency(.$data))
      freq <- if (length(unique(freq)) == 1) unique(freq) else 1
    }
    covm_naive <- .compute_naive_cov(obs, freq = freq, criterion = criterion)
        
    # Identify optimal prior parameters via LOO-CV
    loo_cv = multi_log_score_optimization(res, covm_naive)
    nu_prior = loo_cv$optimal_nu
    Psi_prior = (nu_prior - n_tot - 1) * covm_naive
  }

  # For all horizon steps ahead, apply independently
  fc_dist <- lapply(base_forecast_h, function(base_forecasts) {
    
    # Compute posterior parameters if they are not directly specified
    var_h <- distributional::variance(base_forecasts)
    W_h <- diag(sqrt(var_h))%*%R1%*%t(diag(sqrt(var_h)))
    W_h <- (1 - l_shr)*W_h + l_shr*diag(diag(W_h))
    Psi_post = Psi_prior + nrow(res) * W_h
    nu_post = nu_prior + nrow(res)
    
    # Extrapolate point forecast
    base_fc_mean = map_dbl(base_forecasts, mean)
    out = .core_reconc_t(
      A = A,
      base_fc_mean = base_fc_mean,
      Psi_post = Psi_post,
      nu_post = nu_post,
      return_upper = TRUE,
      return_parameters = FALSE
    )
    
    # Return the distributional Student-t distribution
    return(dist_student_t(
      df = out$bottom_rec_df,
      mu = c(out$upper_rec_mean, out$bottom_rec_mean),
      # S: to check whether the following has to be scaled through sqrt
      sigma = sqrt(c(diag(out$upper_rec_scale_matrix), diag(out$bottom_rec_scale_matrix)))
    ))
  })
  # END REWRITE OF bayesRecon_t
  
  # Fable needs the horizon and models in a different format
  # Invert horizon <-> model.
  fc_dist <- transpose_vec(fc_dist)
  
  # t returns the upper/bottom ordering with upper on top and bottom below
  # fable takes in input series in any arbitrary position so we need to invert back
  # Invert <A/B> smat ordering to arbitrary key_data order
  fc_dist <- fc_dist[order(c(upr_ts, btm_ts[btm_idx]))]
  get_output_fc(fc, fc_dist, point_forecast)
}

data("swiss_tourism", package = "bayesRecon")
saveRDS(swiss_tourism, file = "tests/testthat/data/swiss_tourism.rds")


data("extr_mkt_events", package = "bayesRecon")
saveRDS(extr_mkt_events, file = "tests/testthat/data/extreme_mkt_events.rds")

#' @title Probabilistic reconciliation of mixed hierarchies
#'
#' @description
#'
#' `bayesRecon_MixCond` specifies Mixed-Conditioning (Mix-Cond) reconciliation for use within 
#' `reconcile()`. The method uses importance sampling to draw samples from the reconciled
#' forecast distribution, obtained via conditioning, in the case of a mixed hierarchy.
#' 
#' `bayesRecon_TDcond` specifies Top-Down Conditioning reconciliation for use within 
#' `reconcile()`. The method uses a top-down conditioning algorithm: first, upper base forecasts 
#' are reconciled via conditioning using only the hierarchical constraints between the
#' upper; then, the bottom distributions are updated via a probabilistic top-down procedure.
#' 
#' Reconciliation is performed when `forecast()` is called on the resulting model.
#' Marginal reconciled forecasts follow a sample distribution.
#'
#'
#' @param models A list of fitted models to reconcile.
#' @param n_samples Number of samples to draw from the reconciled distribution.
#' @param suppress_warnings If `TRUE`, suppress warnings from reconciliation.
#' 
#' @return A model specification of class \code{"bayesRecon_MixCond"} or 
#'   \code{"bayesRecon_TDcond"} (inheriting from \code{"mdl_lst"}) 
#'   to be passed to \code{\link[fabletools]{reconcile}}. The reconciliation is
#'   performed when \code{\link[fabletools]{forecast}} is called on
#'   the resulting mable (a model table). The forecast output is a \code{fable} object
#'   containing reconciled probabilistic forecasts represented as
#'   sample distributions (\code{\link[distributional]{dist_sample}}).
#' 
#' @examples
#' library(tsibble)
#' library(dplyr)
#' library(tibble)
#' library(fable)
#' library(fabletools)
#' 
#' 
#' # Mixed hierarchy with integer-valued bottom series and one upper aggregate.
#' # Two low-rate Poisson count series at the bottom; their sum forms the
#' # continuous-looking upper aggregate. 
#' set.seed(42)
#' n <- 60
#' idx <- tsibble::yearmonth("2019 Jan") + 0:(n - 1)
#' 
#' counts <- dplyr::bind_rows(
#'    tibble::tibble(Month = idx, Item = "A", Sales = rpois(n, lambda = 3)),
#'    tibble::tibble(Month = idx, Item = "B", Sales = rpois(n, lambda = 5))) |>
#'    tsibble::as_tsibble(index = Month, key = Item) |>
#'    fabletools::aggregate_key(Item, Sales = sum(Sales))
#'    
#' # Fit a base model on every level, then reconcile via MixCond:
#' # the upper (aggregated) forecast is treated as Gaussian, while the
#' # bottom-level forecast samples are treated as discrete via importance
#' # sampling.
#' fit <- counts |>
#' model(base = ETS(Sales)) |>
#' reconcile(mc = bayesRecon_MixCond(base))
#' 
#' 
#' # Alternative reconciliation via TDcond
#' fit_TDcond <- counts |>
#' model(base = ETS(Sales)) |>
#' reconcile(mc = bayesRecon_TDcond(base))
#' 
#' 
#' fc <- forecast(fit, h = 2)
#' fc
#' 
#' 
#' @references
#' Zambon, L., Azzimonti, D., Rubattu, N., Corani, G. (2024).
#' *Probabilistic reconciliation of mixed-type hierarchical time series*.
#' Proceedings of the Fortieth Conference on Uncertainty in Artificial Intelligence,
#' PMLR 244:4078-4095. <https://proceedings.mlr.press/v244/zambon24a.html>.
#'
#' @seealso [fabletools::reconcile()], [fabletools::aggregate_key()], 
#' [bayesRecon_BUIS()], [bayesRecon::reconc_MixCond()], [bayesRecon::reconc_TDcond()]
#'
#' @name bayesRecon_MixCond
 
#' @rdname bayesRecon_MixCond
#'
#' @export
bayesRecon_MixCond <- function(models, n_samples = 1000, suppress_warnings = TRUE) {
  structure(models, class = c("lst_bayesRecon_MixCond", "mdl_lst", "list"),
            n_samples = n_samples, suppress_warnings = suppress_warnings)
}

#' forecast.lst_bayesRecon_MixCond
#'
#' Produces probabilistic forecasts reconciled via mixed hierarchy conditioning.
#' This method handles mixed hierarchies (with cross-constraints) by applying
#' conditioning-based reconciliation to draw samples from the reconciled distributions.
#'
#' @importFrom fabletools forecast distribution_var
#' @importFrom distributional generate dist_sample
#' @importFrom bayesRecon .core_reconc_MixCond schaferStrimmer_cov
#'
#' @method forecast lst_bayesRecon_MixCond
#'
#' @param object An object of class `lst_bayesRecon_MixCond` containing fitted models.
#' @param key_data A keyed data frame from `fabletools`.
#' @param point_forecast A list of point forecast functions (default: `list(.mean = mean)`).
#' @param new_data Optional new data for forecasting (not currently used).
#' @param ... Additional arguments passed to other methods.
#'
#' @return A fable object with MixCond-reconciled distributions and point forecasts.
#'
#' @keywords internal
#' @export
forecast.lst_bayesRecon_MixCond <- function(
  object,
  key_data,
  point_forecast = list(.mean = mean),
  new_data = NULL,
  ...
) {
  # Take models from fabletools, and prepare for BUIS
  # Produce the structural matrix from the key_data structure
  S <- fabletools::coherent_smat(key_data)
  
  # applies the next method ("mdl_lst", in class definition above) to extract the fitted models.
  fc <- NextMethod()
  
  # Series of lapply to extract the parameters of the distribution
  fc_dist <- lapply(fc, function(x) x[[distribution_var(x)]])
  
  ##### START OUR REWRITE OF reconc_MixCond() with distributional
  hier <- get_hier(S, fc_dist)
  A <- hier$A
  base_forecast_h <- hier$base_forecast_h
  upr_ts <- hier$upr_ts
  btm_ts <- hier$btm_ts
  btm_idx <- hier$btm_idx
  n_upr <- hier$n_upr
  
  # Extrapolate additional parameters from the object attributes
  n_samples <- attr(object, "n_samples")
  suppress_warnings <- attr(object, "suppress_warnings")
  
  # Compute upper sample covariance, drop rows containing nans
  res_upr <- get_residuals(object, upr_ts, btm_ts, btm_idx, n_upr, "upper")
  if (n_upr == 1){
    upr_covm <- matrix(crossprod(res_upr)/nrow(res_upr))
  } else {
    upr_covm <- schaferStrimmer_cov(res_upr)$shrink_cov
  }
  
  # For all horizon steps ahead, apply independently
  fc_dist <- lapply(base_forecast_h, function(base_forecasts) {
    
    # Save upper point forecast vector and bottom samples matrix
    mean_upper <- base_forecasts[seq_len(n_upr)] |> mean()
    B <- base_forecasts[-seq_len(n_upr)] |> generate(times = n_samples) |> do.call(what=cbind)
    
    out <- .core_reconc_MixCond(
      A = A,
      B = B,
      mean_upper = mean_upper,
      cov_upper = upr_covm,
      num_samples = n_samples,
      return_type = "samples",
      suppress_warnings = suppress_warnings
    )
    
    # Return reconciled samples as a distributional object
    Y_reconc = rbind(out$upper_rec_samples, out$bottom_rec_samples)
    return(dist_sample(split(Y_reconc, row(Y_reconc))))
  })
  ###### END REWRITE

  # Fable needs the horizon and models in a different format
  # Invert horizon <-> model. 
  fc_dist <- transpose_vec(fc_dist)

  # MixCond returns the upper/bottom ordering with upper on top and bottom below
  # fable takes in input series in any arbitrary position so we need to invert back
  # Invert <A/B> smat ordering to arbitrary key_data order
  fc_dist <- fc_dist[order(c(upr_ts, btm_ts[btm_idx]))]
  get_output_fc(fc, fc_dist, point_forecast)
}

#' Transpose a list of vectors by position
#'
#' Takes a list of equal-length vectors and returns a list where each element
#' contains the values at a given position across the input list.
#'
#' @details
#' This helper function is useful for converting between horizon-indexed and
#' model-indexed forecasts in hierarchical reconciliation.
#'
#' @param .l A list of vectors with the same length.
#'
#' @return A list of vectors with length equal to the input vector length.
#'
#' @keywords internal
#' @noRd
#' @importFrom vctrs vec_c vec_slice
#' @importFrom rlang abort enquo
transpose_vec <- function(.l) {
  result <- lapply(seq_along(.l[[1]]), function(i) {
    do.call(vec_c, lapply(.l, vec_slice, i))
  })
  return(result)
}

#' Extract hierarchy metadata and base forecasts
#'
#' Computes indices for upper and bottom series, extracts the hierarchy
#' submatrix, and reorders base forecast distributions for bayesRecon.
#'
#' @param S A summing matrix for the hierarchy.
#' @param fc_dist A list of base forecast distributions, one per series.
#'
#' @return A list containing:
#'   \\item{upr_ts}{Indices of upper-level series.}
#'   \\item{btm_ts}{Indices of bottom-level series.}
#'   \\item{A}{The constraint matrix for upper-level reconciliation.}
#'   \\item{btm_idx}{Reordering indices for bottom series.}
#'   \\item{base_forecast_h}{Reordered and transposed distributions.}
#'   \\item{n_upr}{Number of upper-level series.}
#'   \\item{n_btm}{Number of bottom-level series.}
#'   \\item{n_tot}{Total number of series.}
#'
#' @keywords internal
#' @noRd
get_hier <- function(S, fc_dist){
  upr_ts <- which(rowSums(S) > 1)
  btm_ts <- which(rowSums(S) == 1)
  A <- S[rowSums(S) > 1, , drop = FALSE]
  btm_idx <- apply(S[btm_ts, , drop = FALSE], 1, \(x) which(as.logical(x)))
  base_forecast_h <- transpose_vec(fc_dist[c(upr_ts, btm_ts[btm_idx])])
  n_upr <- nrow(A)
  n_btm <- ncol(A)
  n_tot <- n_upr + n_btm
  return(list(
    upr_ts = upr_ts,
    btm_ts = btm_ts,
    A = A,
    btm_idx = btm_idx,
    base_forecast_h = base_forecast_h,
    n_upr = n_upr,
    n_btm = n_btm,
    n_tot = n_tot
  ))
}


#' Assemble fable-compatible forecast output
#'
#' Inserts reconciled distributions and computed point forecasts into
#' the fable forecast object.
#'
#' @param fc List of forecast objects.
#' @param fc_dist List of distributions to insert.
#' @param point_forecast Point forecast spec passed to fabletools.
#'
#' @return A list of updated forecast objects with reconciled distributions
#'   and point forecasts inserted.
#'
#' @importFrom purrr map2
#' @importFrom fabletools distribution_var
#'
#' @keywords internal
#' @noRd
get_output_fc <- function(fc, fc_dist, point_forecast){
  # The code below is Mitch magic that makes the returned object compatible with fable pipeline
  # you can copy paste this in other functions
  # In the next iteration of fable this will become a proper function 
  # (or it won't be needed anymore because it will be handled outside of the reconcile functions)
  out_fc <- map2(fc, fc_dist, function(fc, dist) {
    dimnames(dist) <- dimnames(fc[[distribution_var(fc)]])
    fc[[distribution_var(fc)]] <- dist
    point_fc <- compute_point_forecasts(dist, point_forecast)
    fc[names(point_fc)] <- point_fc
    fc
  })
  return(out_fc)
}


#' Extract residuals for hierarchy levels
#'
#' Computes residuals from model objects and optionally returns a subset
#' corresponding to upper-level, bottom-level, or all series in the hierarchy.
#'
#' @param object List-like collection of model objects.
#' @param upr_ts Integer indices for upper series.
#' @param btm_ts Integer indices for bottom series.
#' @param btm_idx Integer order indices for bottom series.
#' @param level Which level of residuals to return: "all", "upper", or "bottom".
#'
#' @return A numeric matrix of residuals with complete cases only.
#' @importFrom purrr map map_dbl reduce exec
#' @importFrom dplyr full_join
#' @importFrom stats complete.cases residuals
#' @importFrom tsibble index_var
#' @keywords internal
#' @noRd
get_residuals <- function(object, upr_ts, btm_ts, btm_idx, n_upr, 
                          level=c("all","upper","bottom")){
  level <- match.arg(level)
  # Compute sample covariance
  res <- map(
    object[c(upr_ts, btm_ts[btm_idx])], 
    function(x, ...) residuals(x, ...), type = "response")
  if(length(unique(map_dbl(res, nrow))) > 1){
    # Join residuals by index #199
    res <- unname(as.matrix(reduce(res, full_join, by = index_var(res[[1]]))[,-1]))
  } else {
    res <- matrix(exec(c, !!!map(res, `[[`, 2)), ncol = length(object))
  }
  # select the level of residuals to return
  if (level == "upper") {
    res <- res[, seq_len(n_upr), drop = FALSE]
  } else if (level == "bottom") {
    res <- res[, -seq_len(n_upr), drop = FALSE]
  }
  # Drop lines with NAs
  res <- res[complete.cases(res), , drop = FALSE]
  return(res)
}

  
#' Copied from fabletools/R/utils.R
#' @keywords internal
#' @noRd
calc <- function(f, ...){
  f(...)
}

#' Copied from fabletools/R/forecast.R
#' @importFrom purrr map
#' @keywords internal
#' @noRd
compute_point_forecasts <- function(distribution, measures){
  map(measures, calc, distribution)
}

STATICNB <- function(formula, ...) {
  staticnb_model <- fabletools::new_model_class(
    "staticnb",
    train = train_staticnb,
    specials = fabletools::new_specials(),
    check = all_tsbl_checks
  )
  fabletools::new_model_definition(staticnb_model, !!rlang::enquo(formula), ...)
}

train_staticnb <- function(.data, specials, ...) {
  if (length(tsibble::measured_vars(.data)) != 1) {
    cli::cli_abort("Only univariate responses are supported by STATICNB.")
  }
  y <- unclass(.data)[[tsibble::measured_vars(.data)]]
  mu <- mean(y, na.rm = TRUE)
  mu <- validate_mu(mu, context = "train")

  structure(
    list(
      par = list(
        term = c("mean", "size", "prob"),
        estimate = c(mu, mu, 0.5)
      ),
      .mean = mu,
      .size = mu,
      .prob = 0.5,
      .fitted = rep(mu, length(y)),
      .residuals = y - mu
    ),
    class = "staticnb"
  )
}

#' @inherit forecast.ARIMA
forecast.staticnb <- function(object, new_data, ...) {
  h <- nrow(new_data)
  mu <- validate_mu(object$.mean, context = "forecast")
  prob <- object$.prob
  distributional::dist_negative_binomial(size = rep(mu, h), prob = rep(prob, h))
}

#' @inherit fitted.ARIMA
fitted.staticnb <- function(object, ...) {
    object[[".fitted"]]
}

#' @inherit tidy.ARIMA
residuals.staticnb <- function(object, ...) {
    object[[".residuals"]]
}

validate_mu <- function(mu, context) {
  if (length(mu) != 1L) {
    rlang::abort(paste0("STATICNB: ", context, " mean is not a scalar."))
  }
  if (is.na(mu) || !is.finite(mu)) {
    rlang::abort(paste0("STATICNB: ", context, " mean is NA/NaN/Inf. Check response values."))
  }
  if (mu < 0) {
    rlang::abort(paste0("STATICNB: ", context, " mean is negative. Check response values."))
  }
  mu
}



check_gaps <- function(x) {
  if (any(tsibble::has_gaps(x)[[".gaps"]])) {
    cli::cli_abort(sprintf("%s contains implicit gaps in time. You should check your data and convert implicit gaps into explicit missing values using `tsibble::fill_gaps()` if required.", deparse(substitute(x))))
  }
}

check_regular <- function(x) {
  if (!tsibble::is_regular(x)) {
    cli::cli_abort(sprintf("%s is an irregular time series, which this model does not support. You should consider if your data can be made regular, and use `tsibble::update_tsibble(%s, regular = TRUE)` if appropriate.", deparse(substitute(x)), deparse(substitute(x))))
  }
}

check_ordered <- function(x) {
  if (!tsibble::is_ordered(x)) {
    cli::cli_abort(sprintf(
      "%s is an unordered time series. To use this model, you first must sort the data in time order using `dplyr::arrange(%s, %s)`",
      deparse(substitute(x)), paste(c(deparse(substitute(x)), tsibble::key_vars(x)), collapse = ", "), tsibble::index_var(x)
    ))
  }
}
all_tsbl_checks <- function(.data) {
  check_gaps(.data)
  check_regular(.data)
  check_ordered(.data)
  if (NROW(.data) == 0) {
    cli::cli_abort("There is no data to model. Please provide a dataset with at least one observation.")
  }
}

# Ensure S3 methods are registered after fabletools is loaded
if (requireNamespace("fabletools", quietly = TRUE)) {
  tryCatch({
    registerS3method("forecast", "staticnb", forecast.staticnb, envir = asNamespace("fabletools"))
    registerS3method("fitted", "staticnb", fitted.staticnb)
    registerS3method("residuals", "staticnb", residuals.staticnb)
  }, error = function(e) {
    # If registration fails, methods should still work via naming convention
    invisible(NULL)
  })
}

#' BUIS for probabilistic reconciliation of forecasts via conditioning
#' 
#' @description
#'
#' Specifies Bottom-Up Importance Sampling (BUIS) reconciliation for use within 
#' `reconcile()`. The method uses the Bottom-Up Importance Sampling algorithm 
#' to draw samples from the reconciled forecast distribution, obtained via conditioning.
#' Reconciliation is performed when `forecast()` is called on the resulting model.
#' Marginal reconciled forecasts follow a sample distribution.
#'
#' @param models A list of fitted models to reconcile.
#' @param n_samples Number of samples to draw from the reconciled distribution.
#' 
#' @return A model specification of class \code{"bayesRecon_BUIS"}
#'   (inheriting from \code{"mdl_lst"}) to be passed to
#'   \code{\link[fabletools]{reconcile}}. The reconciliation is
#'   performed when \code{\link[fabletools]{forecast}} is called on
#'   the resulting mable (a model table). The forecast output is a \code{fable} object
#'   containing reconciled probabilistic forecasts represented as
#'   sample distributions (\code{\link[distributional]{dist_sample}}).
#'   
#' @examples 
#' library(fable)
#' library(fabletools)
#' library(tsibble)
#' library(fable.bayesRecon)
#' 
#' # Small hierarchical example using tourism data
#' tourism_small <- tsibble::tourism |>
#' dplyr::filter(Region == "Melbourne") |> 
#' fabletools::aggregate_key(Purpose, Trips = sum(Trips))
#' 
#' fit <- tourism_small |> 
#' model(base = ETS(Trips ~ trend("A") + season("A"))) |>
#' reconcile(recon = bayesRecon_BUIS(base, n_samples = 300))
#' 
#' fc <- forecast(fit, h = "2 years")
#' fc
#'
#' @references
#' Zambon, L., Azzimonti, D. & Corani, G. (2024).
#' *Efficient probabilistic reconciliation of forecasts for real-valued and count time series*.
#' Statistics and Computing 34 (1), 21.
#' \doi{10.1007/s11222-023-10343-y}.
#'
#' @seealso [fabletools::reconcile()], [fabletools::aggregate_key()], [bayesRecon_MixCond()], [bayesRecon::reconc_BUIS()]
#'
#' @export
bayesRecon_BUIS <- function(models, n_samples = 1000) {
  structure(models, class = c("lst_bayesRecon_BUIS", "mdl_lst", "list"),
            n_samples = n_samples)
}

#' forecast.lst_bayesRecon_BUIS
#'
#' Produces probabilistic forecasts reconciled via Bottom-Up Importance Sampling (BUIS).
#' This method samples from bottom-level distributions, computes upper-level paths via
#' aggregation, and reweights samples according to upper-level forecast densities.
#'
#' @importFrom fabletools forecast distribution_var
#' @importFrom distributional generate dist_sample support
#' @importFrom stats density
#' @importFrom bayesRecon .check_hierarchical .core_reconc_BUIS .get_HG
#'
#' @method forecast lst_bayesRecon_BUIS
#'
#' @param object An object of class `lst_bayesRecon_BUIS` containing fitted models.
#' @param key_data A keyed data frame from `fabletools`.
#' @param point_forecast A list of point forecast functions (default: `list(.mean = mean)`).
#' @param new_data Optional new data for forecasting (not currently used).
#' @param ... Additional arguments passed to other methods.
#'
#' @return A fable object with BUIS-reconciled distributions and point forecasts.
#'
#' @keywords internal
#' @export
forecast.lst_bayesRecon_BUIS <- function(
  object,
  key_data,
  point_forecast = list(.mean = mean),
  new_data = NULL,
  ...
) {
  # Take models from fabletools, and prepare for BUIS
  # Produce the structural matrix from the key_data structure
  S <- fabletools::coherent_smat(key_data)
  
  # core_reconc_BUIS <- getFromNamespace(".core_reconc_BUIS", "bayesRecon")
  # get_HG <- getFromNamespace(".get_HG", "bayesRecon")
  
  # applies the next method ("mdl_lst", in class definition above) to extract the fitted models.
  fc <- NextMethod()
  
  # Series of lapply to extract the parameters of the distribution
  fc_dist <- lapply(fc, function(x) x[[distribution_var(x)]])
  
  ##### START OUR REWRITE OF reconc_BUIS() with distributional
  hier <- get_hier(S, fc_dist)
  A <- hier$A
  base_forecast_h <- hier$base_forecast_h
  n_upr <- hier$n_upr
  n_btm <- hier$n_btm
  upr_ts <- hier$upr_ts
  btm_ts <- hier$btm_ts
  btm_idx <- hier$btm_idx
  # browser()
  # Check that if bottom are continuous, then all forecasts are
  for (h in seq_along(base_forecast_h)) {
    supp <- base_forecast_h[[h]] |> support()
    if (
      any(supp[-seq_len(n_upr)] |> format()   %in% c("R", "R+")) && 
        !all(supp[seq_len(n_upr)] |> format() %in% c("R", "R+"))
      ) {
      stop("If bottom forecasts are continuous, upper forecasts must be continuous too.")
    }
  }
  
  # Get a core parameter from the object attributes
  n_samples <- attr(object, "n_samples")
  
  # For all horizon steps ahead, apply independently
  fc_dist <- lapply(base_forecast_h, function(base_forecasts) {
    
    # save two lists of upper and bottom fc
    upper_fc <- base_forecasts[seq_len(n_upr)]
    bottom_fc <- base_forecasts[-seq_len(n_upr)]

    # sample from bottom fc and make it a matrix
    B <- bottom_fc |> generate(times = n_samples) |> do.call(what=cbind)

    is.hier = .check_hierarchical(A)
    if(is.hier){
      H <- A
      G <- NULL
      upp_base_H = upper_fc
      upp_base_G = NULL
      in_typeH = NULL
      distr_H  = NULL
      in_typeG = NULL
      distr_G  = NULL
    }else{
      get_HG.res = .get_HG(A, upper_fc, rep(0, n_upr), rep(0, n_upr))
      H = get_HG.res$H
      upp_base_H = get_HG.res$Hv
      G = get_HG.res$G
      upp_base_G = get_HG.res$Gv
      in_typeH = NULL
      distr_H  = NULL
      in_typeG = NULL
      distr_G  = NULL
    }

    .comp_w_distributional <- function(b, u,    
                                       in_type_ = NULL, 
                                       distr_ = NULL){
      return(density(u, as.numeric(b))[[1L]])
    }

    out = .core_reconc_BUIS(A = A, H = H, G = G, B = B,
                          upper_base_fc_H = upp_base_H,
                          in_typeH = in_typeH, distr_H = distr_H,
                          upper_base_fc_G = upp_base_G,
                          in_typeG = in_typeG, distr_G = distr_G,
                          .comp_w = .comp_w_distributional,
                          return_upper = TRUE
                          )

    Y_reconc = rbind(out$upper_rec_samples, out$bottom_rec_samples)
    return(dist_sample(split(Y_reconc, row(Y_reconc))))
    ### END REWRITE
  })

  # Fable needs the horizon and models in a different format
  # Invert horizon <-> model. 
  fc_dist <- transpose_vec(fc_dist)

  # BUIS returns the upper/bottom ordering with upper on top and bottom below
  # fable takes in input series in any arbitrary position so we need to invert back
  # Invert <A/B> smat ordering to arbitrary key_data order
  fc_dist <- fc_dist[order(c(upr_ts, btm_ts[btm_idx]))]
  get_output_fc(fc, fc_dist, point_forecast)
}

tiny_m5 <- readRDS(testthat::test_path("data", "tiny_m5.rds"))

# This function filters uperr or bottom series so that a model can be applied to them
hier_filter <- function(data, level = c("upper", "bottom")) {
  level <- match.arg(level)
  key_cols <- tsibble::key_vars(data)
  if (level == "upper") {
    data <- data |> dplyr::filter(dplyr::if_any(dplyr::all_of(key_cols), fabletools::is_aggregated))
  } else if (level == "bottom") {
    data <- data |> dplyr::filter(!dplyr::if_any(dplyr::all_of(key_cols), fabletools::is_aggregated))
  } else {
    stop("Invalid level specified. Use 'upper' or 'bottom'.")
  }
  return(data)
}


tourism_melbourne <- function() {
  data <- tsibble::tourism |>
    dplyr::filter(Region == "Melbourne") |> 
    fabletools::aggregate_key(Purpose, Trips = sum(Trips))

  fit <- data |> 
    dplyr::filter(Quarter < tsibble::yearquarter("2015 Q1")) |> 
    fabletools::model(base = fable::ETS(Trips ~ trend("A") + season("A")))
  return(fit)
}


tourism_2purposes2states <- function() {
  data <- tsibble::tourism |>
    dplyr::filter(Purpose %in% c("Business", "Holiday")) |>
    dplyr::filter(State %in% c("Northern Territory", "Western Australia")) |>
    fabletools::aggregate_key(Purpose * (State / Region), Trips = sum(Trips))
  
  fit <- data |> 
    dplyr::filter(Quarter < tsibble::yearquarter("2015 Q1")) |> 
    fabletools::model(base = fable::THETA(Trips))
  return(fit)
}


m5_ca1 <- function() {
  data <- tiny_m5 |>
    dplyr::filter(store_id == "CA_1") |>
    tsibble::tsibble(index = date, key = c(item_id, dept_id, cat_id)) |>
    fabletools::aggregate_key(cat_id / dept_id / item_id, value = sum(value))

  fit_upper <- data |> 
    dplyr::filter(date < as.Date("2016-04-01")) |> 
    hier_filter("upper") |>
    fabletools::model(base = fable::THETA(value))
  fit_bottom <- data |> 
    dplyr::filter(date < as.Date("2016-04-01")) |> 
    hier_filter("bottom") |>
    fabletools::model(base = STATICNB(value))
  fit <- dplyr::bind_rows(fit_upper, fit_bottom)
  return(fit)
}


m5_foods1046 <- function() {
  data <- tiny_m5 |>
    dplyr::filter(item_id == "FOODS_1_046") |>
    tsibble::tsibble(index = date, key = c(store_id, state_id)) |>
    fabletools::aggregate_key(state_id / store_id, value = sum(value))

  fit_upper <- data |> 
    dplyr::filter(date < as.Date("2016-04-01")) |> 
    hier_filter("upper") |>
    fabletools::model(base = fable::ETS(value ~ trend("A")))
  fit_bottom <- data |> 
    dplyr::filter(date < as.Date("2016-04-01")) |> 
    hier_filter("bottom") |>
    fabletools::model(base = STATICNB(value))
  fit <- dplyr::bind_rows(fit_upper, fit_bottom)
  return(fit)
}


m5_stores <- function() {
  data <- tiny_m5 |> 
    tsibble::tsibble(index = date, key = c(item_id, store_id, state_id)) |>
    fabletools::aggregate_key(state_id / store_id, value = sum(value))
  
  fit <- data |>
    dplyr::filter(date < as.Date("2016-04-01")) |> 
    fabletools::model(base = fable::SNAIVE(value, period = 7))
  return(fit)
}


pedestrian_all <- function() {
  data <- tsibble::pedestrian |>
    dplyr::filter(Date_Time < as.Date("2015-04-01")) |>
    fabletools::aggregate_key(Sensor, Count = sum(Count))
  
  fit_upper <- data |>
    hier_filter("upper") |>
    fabletools::model(base = fable::ETS(Count ~ trend("A") + season("A")))
  fit_bottom <- data |>
    hier_filter("bottom") |>
    fabletools::model(base = STATICNB(Count))
  fit <- dplyr::bind_rows(fit_upper, fit_bottom)
  return(fit)
}

carparts_100 <- function(){
  data <- tsibbledata::monash_forecasting_repository(4656021) |> 
    dplyr::mutate(start_timestamp = tsibble::yearmonth(start_timestamp)) |>
    dplyr::filter(series_name %in% paste0("T", 42 + 0:99)) |> 
    dplyr::filter(start_timestamp < tsibble::yearmonth("2001 Oct")) |>
    fabletools::aggregate_key(series_name, value = sum(value))
  
  fit_upper <- data |>
    hier_filter("upper") |>
    fabletools::model(base = fable::SNAIVE(value, period = 12))
  fit_bottom <- data |>
    hier_filter("bottom") |>
    fabletools::model(base = STATICNB(value))
  fit <- dplyr::bind_rows(fit_upper, fit_bottom)
}


carparts_5 <- function(){
  data <- tsibbledata::monash_forecasting_repository(4656021) |> 
    dplyr::mutate(start_timestamp = tsibble::yearmonth(start_timestamp)) |>
    dplyr::filter(series_name %in% paste0("T", 42 + 0:4)) |> 
    dplyr::filter(start_timestamp < tsibble::yearmonth("2001 Oct")) |>
    fabletools::aggregate_key(series_name, value = sum(value))
  
  fit_upper <- data |>
    hier_filter("upper") |>
    fabletools::model(base = STATICNB(value))
  fit_bottom <- data |>
    hier_filter("bottom") |>
    fabletools::model(base = fable::SNAIVE(value, period = 12))
  fit <- dplyr::bind_rows(fit_upper, fit_bottom)
  return(fit)
}


swiss_tourism_all <- function(){
  data <- bayesRecon::swiss_tourism$ts[, -1, drop = FALSE] |>
    tsibble::as_tsibble() |>
    dplyr::rename(Month = index, Canton = key, Tourists = value) |> 
    fabletools::aggregate_key(Canton, Tourists = sum(Tourists))
  fit <- data |>
    dplyr::filter(Month < tsibble::yearmonth("2024 Jan")) |> 
    fabletools::model(base = fable::ETS(Tourists ~ trend("A") + season("A")))
  return(fit)
}

extr_mkt_events_all <- function(){
  data <- bayesRecon::extr_mkt_events[, -1, drop = FALSE] |>
    tsibble::as_tsibble() |>
    dplyr::rename(MarketDay = index, Sector = key, Events = value) |> 
    fabletools::aggregate_key(Sector, Events = sum(Events))
  fit <- data |>
    dplyr::filter(MarketDay < 3499) |> 
    fabletools::model(base = STATICNB(Events))
  return(fit)
}


distributions <- function(type = c("nonnegative_integer", "nonnegative_continuous", "real_valued")){
  type <- match.arg(type)
  distr <- distributional::dist_poisson(2.)
  if (type == "nonnegative_integer") {
    distr <- c(distr, distributional::dist_negative_binomial(3., 0.6))
  } else if (type == "nonnegative_continuous") {
    distr <- c(distr, distributional::dist_gamma(2., 1.))
  } else if (type == "real_valued") {
    distr <- c(distr, distributional::dist_normal(0., 1.))
  } else {
    stop("Invalid type specified.")
  }
  return(distr)
}

testthat::test_that("bayesRecon_MixCond runs reconciliation and produces valid output", {
  testthat::skip_if_not_installed("fable")
  testthat::skip_if_not_installed("fabletools")
  testthat::skip_if_not_installed("tsibble")
  testthat::skip_if_not_installed("distributional")

  # Test 1
  testthat::expect_no_error({
    fc1 <- carparts_100() |>
      fabletools::reconcile(mixcond = fable.bayesRecon::bayesRecon_MixCond(base, 
                                                                           n_samples = 500)) |>
      fabletools::forecast(h = 6) |> 
      dplyr::filter(.model == "mixcond")
    distr1 <- fc1[[fabletools::distribution_var(fc1)]]
  })
  
  # Test 2
  testthat::expect_no_error({
    fc2 <- m5_foods1046() |>
      fabletools::reconcile(mixcond = fable.bayesRecon::bayesRecon_MixCond(base)) |>
      fabletools::forecast(h = 28) |> 
      dplyr::filter(.model == "mixcond")
    distr2 <- fc2[[fabletools::distribution_var(fc2)]]
  })
  
  # Test 3
  testthat::expect_no_error({
    fc3 <- pedestrian_all() |>
      fabletools::reconcile(mixcond = fable.bayesRecon::bayesRecon_MixCond(base, 
                                                                           n_samples = 2000)) |>
      fabletools::forecast(h = "24 hours") |> 
      dplyr::filter(.model == "mixcond")
    distr3 <- fc3[[fabletools::distribution_var(fc3)]]
  })
  
  # Check the class of the forecasts
  testthat::expect_s3_class(fc1, c("tbl_df", "tbl"))
  testthat::expect_s3_class(fc2, c("tbl_df", "tbl"))
  testthat::expect_s3_class(fc3, c("tbl_df", "tbl"))
  
  # Check the existence of model and mean columns
  testthat::expect_true(all(c(".model", ".mean") %in% names(fc1)))
  testthat::expect_true(all(c(".model", ".mean") %in% names(fc2)))
  testthat::expect_true(all(c(".model", ".mean") %in% names(fc3)))
  
  # Check the existence of point forecasts (not NA)
  testthat::expect_true(all(fc1$.mean |> is.finite()))
  testthat::expect_true(all(fc2$.mean |> is.finite()))
  testthat::expect_true(all(fc3$.mean |> is.finite()))
  
  # Check that the distribution is a distributional object
  testthat::expect_true(distr1 |> distributional::is_distribution())
  testthat::expect_true(distr2 |> distributional::is_distribution())
  testthat::expect_true(distr3 |> distributional::is_distribution())
  
  # Test that the distirbution is a sample distribution
  testthat::expect_true(all(distr1 |> stats::family() == "sample"))
  testthat::expect_true(all(distr2 |> stats::family() == "sample"))
  testthat::expect_true(all(distr3 |> stats::family() == "sample"))
})

testthat::test_that("bayesRecon_TDcond runs reconciliation and produces valid output", {
  testthat::skip_if_not_installed("fable")
  testthat::skip_if_not_installed("fabletools")
  testthat::skip_if_not_installed("tsibble")
  testthat::skip_if_not_installed("distributional")

  # Test 1
  testthat::expect_no_error({
    fc1 <- carparts_100() |>
      fabletools::reconcile(tdcond = fable.bayesRecon::bayesRecon_TDcond(base, 
                                                                         n_samples = 500)) |>
      fabletools::forecast(h = 6) |> 
      dplyr::filter(.model == "tdcond")
    distr1 <- fc1[[fabletools::distribution_var(fc1)]]
  })
  
  # Test 2
  testthat::expect_no_error({
    fc2 <- m5_foods1046() |>
      fabletools::reconcile(tdcond = fable.bayesRecon::bayesRecon_TDcond(base, 
                                                                         n_samples = 2000)) |>
      fabletools::forecast(h = 28) |> 
      dplyr::filter(.model == "tdcond")
    distr2 <- fc2[[fabletools::distribution_var(fc2)]]
  })
  
  # Test 3 
  testthat::expect_no_error({
    fc3 <- m5_ca1() |>
      fabletools::reconcile(tdcond = fable.bayesRecon::bayesRecon_TDcond(base)) |>
      fabletools::forecast(h = 28) |> 
      dplyr::filter(.model == "tdcond")
    distr3 <- fc3[[fabletools::distribution_var(fc3)]]
  })
  
  # Check the class of the forecasts
  testthat::expect_s3_class(fc1, c("tbl_df", "tbl"))
  testthat::expect_s3_class(fc2, c("tbl_df", "tbl"))
  testthat::expect_s3_class(fc3, c("tbl_df", "tbl"))
  
  # Check the existence of model and mean columns
  testthat::expect_true(all(c(".model", ".mean") %in% names(fc1)))
  testthat::expect_true(all(c(".model", ".mean") %in% names(fc2)))
  testthat::expect_true(all(c(".model", ".mean") %in% names(fc3)))
  
  # Check the existence of point forecasts (not NA)
  testthat::expect_true(all(fc1$.mean |> is.finite()))
  testthat::expect_true(all(fc2$.mean |> is.finite()))
  testthat::expect_true(all(fc3$.mean |> is.finite()))
  
  # Check that the distribution is a distributional object
  testthat::expect_true(distr1 |> distributional::is_distribution())
  testthat::expect_true(distr2 |> distributional::is_distribution())
  testthat::expect_true(distr3 |> distributional::is_distribution())
})


testthat::test_that("make_PMF creates valid PMFs and produces appropriate warnings", {
  
  # Check that the function make_PMF runs without errors
  testthat::expect_no_error({
    pmf1 <- fable.bayesRecon:::make_PMF(distributions("nonnegative_integer"))
  })
  
  # Check that warnings are produced when bottom forecasts are not integer-valued
  testthat::expect_warning(pmf2 <- fable.bayesRecon:::make_PMF(distributions("nonnegative_continuous")))
  testthat::expect_warning(pmf3 <- fable.bayesRecon:::make_PMF(distributions("real_valued"), negative_to_zero = FALSE))
  testthat::expect_warning(pmf4 <- fable.bayesRecon:::make_PMF(distributions("real_valued"), negative_to_zero = TRUE))
  
  # Check that the pfms are a list of numeric vectors
  testthat::expect_true(is.list(pmf1) && all(sapply(pmf1, is.numeric)))
  testthat::expect_true(is.list(pmf2) && all(sapply(pmf2, is.numeric)))
  testthat::expect_true(is.list(pmf3) && all(sapply(pmf3, is.numeric)))
  testthat::expect_true(is.list(pmf4) && all(sapply(pmf4, is.numeric)))
})

testthat::test_that("bayesRecon_t runs reconciliation and produces valid output", {

  # Test 1
  testthat::expect_no_error({
    fc1 <- tourism_melbourne() |>
      fabletools::reconcile(t = fable.bayesRecon::bayesRecon_t(
        base, 
        prior = list(nu = 7, Psi = diag(5) * 3000))) |>
      fabletools::forecast(h = "2 years") |> 
      dplyr::filter(.model == "t")
    distr1 <- fc1[[fabletools::distribution_var(fc1)]]
  })
  
  # Test 2
  testthat::expect_no_error({
    fc2 <- tourism_2purposes2states() |>
      fabletools::reconcile(t = fable.bayesRecon::bayesRecon_t(base)) |>
      fabletools::forecast(h = "2 years") |> 
      dplyr::filter(.model == "t")
    distr2 <- fc2[[fabletools::distribution_var(fc2)]]
  })
  
  # Test 3
  testthat::expect_no_error({
    fc3 <- m5_stores() |>
      fabletools::reconcile(t = fable.bayesRecon::bayesRecon_t(base)) |>
      fabletools::forecast(h = 28) |> 
      dplyr::filter(.model == "t")
    distr3 <- fc3[[fabletools::distribution_var(fc3)]]
  })
  
  # Test 4
  testthat::expect_no_error({
    fc4 <- swiss_tourism_all() |>
      fabletools::reconcile(t = fable.bayesRecon::bayesRecon_t(
        base, 
        prior = list(nu = 50, Psi = 10e6 * (200*diag(27) + matrix(1, 27, 27))))) |>
      fabletools::forecast(h = 6) |> 
      dplyr::filter(.model == "t")
    distr4 <- fc4[[fabletools::distribution_var(fc4)]]
  })
  
  
  # Check the class of the forecasts
  testthat::expect_s3_class(fc1, c("fbl_ts", "tbl_ts"))
  testthat::expect_s3_class(fc2, c("fbl_ts", "tbl_ts"))
  testthat::expect_s3_class(fc3, c("fbl_ts", "tbl_ts"))
  testthat::expect_s3_class(fc4, c("fbl_ts", "tbl_ts"))
  
  # Check the existence of model and mean columns
  testthat::expect_true(all(c(".model", ".mean") %in% names(fc1)))
  testthat::expect_true(all(c(".model", ".mean") %in% names(fc2)))
  testthat::expect_true(all(c(".model", ".mean") %in% names(fc3)))
  testthat::expect_true(all(c(".model", ".mean") %in% names(fc4)))
  
  # Check the existence of point forecasts (not NA)
  testthat::expect_true(all(fc1$.mean |> is.finite()))
  testthat::expect_true(all(fc2$.mean |> is.finite()))
  testthat::expect_true(all(fc3$.mean |> is.finite()))
  testthat::expect_true(all(fc4$.mean |> is.finite()))
  
  # Check that the distribution is a distributional object
  testthat::expect_true(distr1 |> distributional::is_distribution())
  testthat::expect_true(distr2 |> distributional::is_distribution())
  testthat::expect_true(distr3 |> distributional::is_distribution())
  testthat::expect_true(distr4 |> distributional::is_distribution())
  
  # Check that the distribution is a student_t distribution
  testthat::expect_true(all(distr1 |> stats::family() == "student_t"))
  testthat::expect_true(all(distr2 |> stats::family() == "student_t"))
  testthat::expect_true(all(distr3 |> stats::family() == "student_t"))
  testthat::expect_true(all(distr4 |> stats::family() == "student_t"))
  
  # Check that the model fails when base forecasts are not Gaussian
  testthat::expect_error({
    fc5 <- carparts_5() |>
      fabletools::reconcile(t = fable.bayesRecon::bayesRecon_t(base)) |>
      fabletools::forecast(h = 12)
  })
})

#' @rdname bayesRecon_MixCond
#'
#' @export
bayesRecon_TDcond <- function(models, n_samples = 1000, suppress_warnings = TRUE) {
  structure(models, class = c("lst_bayesRecon_TDcond", "mdl_lst", "list"),
            n_samples = n_samples, suppress_warnings = suppress_warnings)
}

#' forecast.lst_bayesRecon_TDcond
#' 
#' Produces probabilistic forecasts reconciled via top-down conditioning.
#' Upper base forecasts are reconciled via conditioning. The recondiled
#' upper forecasts are then propagate to bottom via probabilisitc top-down.
#' 
#' @importFrom fabletools forecast distribution_var
#' @importFrom distributional dist_sample dist_truncated support
#' @importFrom stats density
#' @importFrom vctrs field
#' @importFrom bayesRecon .core_reconc_TDcond schaferStrimmer_cov
#' 
#' @method forecast lst_bayesRecon_TDcond 
#' 
#' @param object An object of class `lst_bayesRecon_MixCond` containing fitted models.
#' @param key_data A keyed data frame from `fabletools`.
#' @param point_forecast A list of point forecast functions (default: `list(.mean = mean)`).
#' @param new_data Optional new data for forecasting (not currently used).
#' @param ... Additional arguments passed to other methods.
#' @param ... extra parameters to be passed on.
#' 
#' @return A fable object with TDcond-reconciled distributions and point forecasts.
#' 
#' @keywords internal
#' @export
forecast.lst_bayesRecon_TDcond <- function(
    object,
    key_data,
    point_forecast = list(.mean = mean),
    new_data = NULL,
    ...
) {
  # Take models from fabletools, and prepare for BUIS
  # Produce the structural matrix from the key_data structure
  S <- fabletools::coherent_smat(key_data)
  
  # applies the next method ("mdl_lst", in class definition above) to extract the fitted models.
  fc <- NextMethod()
  
  # Series of lapply to extract the parameters of the distribution
  fc_dist <- lapply(fc, function(x) x[[distribution_var(x)]])
  
  ##### START OUR REWRITE OF reconc_TDcond() with distributional
  hier <- get_hier(S, fc_dist)
  A <- hier$A
  base_forecast_h <- hier$base_forecast_h
  upr_ts <- hier$upr_ts
  btm_ts <- hier$btm_ts
  btm_idx <- hier$btm_idx
  n_upr <- hier$n_upr
  
  # Extract parameters from the object attributes
  n_samples <- attr(object, "n_samples")
  suppress_warnings <- attr(object, "suppress_warnings")
  
  # Compute upper sample covariance, drop rows containing NANs
  res_upr <- get_residuals(object, upr_ts, btm_ts, btm_idx, n_upr, "upper")
  if (n_upr == 1){
    upr_covm <- matrix(crossprod(res_upr)/nrow(res_upr))
  } else {
    upr_covm <- schaferStrimmer_cov(res_upr)$shrink_cov
  }
  
  # Iterate the reconciliation across horizons
  fc_dist <- lapply(base_forecast_h, function(base_forecasts) {
    
    # Save upper point forecast and bottom PMF
    mean_upper <- base_forecasts[seq_len(n_upr)] |> mean()
    L_pmf <- make_PMF(base_forecasts[-seq_len(n_upr)])

    # Apply the core function from bayesRecon
    out <- .core_reconc_TDcond(
      A = A,
      mean_upper = mean_upper,
      cov_upper = upr_covm,
      L_pmf = L_pmf,
      num_samples =  n_samples,
      return_type = "samples", 
      min_fraction_samples_ok = .5,
      suppress_warnings = suppress_warnings
    )
    
    # Return reconciled samples as a distributional object
    Y_reconc = rbind(out$upper_rec_samples, out$bottom_rec_samples)
    return(dist_sample(split(Y_reconc, row(Y_reconc))))
  })
  # END REWRITE
  
  # Fable needs the horizon and models in a different format
  # Invert horizon <-> model. 
  fc_dist <- transpose_vec(fc_dist)
  
  # MixCond returns the upper/bottom ordering with upper on top and bottom below
  # fable takes in input series in any arbitrary position so we need to invert back
  # Invert <A/B> smat ordering to arbitrary key_data order
  fc_dist <- fc_dist[order(c(upr_ts, btm_ts[btm_idx]))]
  get_output_fc(fc, fc_dist, point_forecast)
}


#' Create a discrete PMF from distributions
#'
#' Converts a set of distributional objects into a list of discrete PMFs used by
#' the top-down conditioning reconciliation. The function enforces nonnegative,
#' integer support by truncating negative supports and rounding non-integer
#' supports via sampling, then estimates the PMF up to the (1 - toll)
#' quantile.
#'
#' @param dist A vector or list of `distributional` objects.
#' @param negative_to_zero Logical, currently unused. Reserved for future use.
#' @param toll Tail probability used to truncate the support when estimating the PMF.
#' @param alpha_smoothing Numeric, currently unused. Reserved for future use.
#'
#' @return A list of numeric vectors, one PMF per input distribution.
#'
#' @importFrom distributional dist_truncated support dist_inflated cdf
#' @importFrom vctrs field
#' @importFrom purrr map2 pmap
#' @importFrom stats density quantile
#'
#' @keywords internal
make_PMF <- function(dist, negative_to_zero = FALSE, toll = 1e-9, alpha_smoothing = 1e-9){
  wm <- character(0)
  supp <- dist |> support()
  # Identify the negatively supported distributions
  neg_lb <- supp |> field("lim") |> vapply(\(x) x[1], numeric(1)) < 0
  if (any(neg_lb)){
    if (negative_to_zero){
      # Identify the mass below zero and make the distribution zero-inflated
      wm <- c(wm, "Negative support corrected via zero-inflation.")
      cdfzero <- dist[neg_lb] |> cdf(-toll)
      dist[neg_lb] <- map2(dist[neg_lb], cdfzero, 
                           \(d, p0) d |> 
                           dist_truncated(0, Inf) |> 
                           dist_inflated(p0, 0))
    } else {
      # Simply truncate the distribution
      wm <- c(wm, "Negative support corrected via truncation.")
      dist[neg_lb] <- dist[neg_lb] |> dist_truncated(0, Inf)
    }
  }
  # Identify real-supported distributions
  int_val <- supp |> format() %in% c("N0", "N+", "Z")
  if (any(!int_val)){
    wm <- c(wm, "Non-integer support obtained via cdf rounding.")
  }
  # Find the upper value for the pmf esitimation
  M <- dist |> quantile(1 - toll) |> ceiling() |> as.numeric()
  pmf <- pmap(list(d = dist, is_int = int_val, m = M), \(d, is_int, m) {
      if (!is_int) {
        # Approximate the pmf via cdf differences
        mass <- d |> cdf(0:m + 0.5) |> 
          (\(x) as.numeric(unlist(x, recursive = TRUE, use.names = FALSE)))() |>
          (\(x) diff(c(0, x)))()
      } else {
        # Simply estimate the pmf
        mass <- d |> density(0:m) |> 
          (\(x) as.numeric(unlist(x, recursive = TRUE, use.names = FALSE)))()
      }
      return(mass)
    }
  )
  # Display warnings if needed
  if (length(wm) > 0) warning(paste(unique(wm), collapse = " "), call. = FALSE)
  return(pmf)
}

---
title: "Forecasts reconciliation with fable.bayesRecon"
output: rmarkdown::html_vignette
bibliography: references.bib
vignette: >
  %\VignetteIndexEntry{Forecasts reconciliation with fable.bayesRecon}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoded{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE, warning = FALSE, message = FALSE)
options(tibble.width = Inf, knitr.kable.NA = "")

# Helper to render fabletools tables as HTML-safe kable
kable <- function(x, ...) {
  x |>
    dplyr::mutate(dplyr::across(
      where(~ inherits(.x, "agg_vec")),
      ~ gsub("<", "&lt;", gsub(">", "&gt;", as.character(.x)))
    )) |>
    knitr::kable(...)
}
```


## Introduction

This vignette shows how to use the forecast reconciliation methods implemented in the `fable.bayesRecon` package.
In particular, we will apply t-Rec method [@carrara2025] to the `swiss_tourism` dataset, which contains monthly counts of nightly stays in Switzerland, aggregated by Canton. 

The method t-Rec is implemented by the function `bayesRecon_t()`, which can be used within the `reconcile()` function of `fabletools` to specify the reconciliation strategy.

First, we load the required packages:

```{r libraries}
library(bayesRecon)
library(dplyr)
library(tsibble)
library(fable)
library(fabletools)
library(fable.bayesRecon)
library(ggplot2)
library(scales)
```


## Data preparation

Let's import the data in the tsibble format. To do so, we follow these steps:

1. Load the `swiss_tourism` dataset from the `bayesRecon` package. This is a list containing the `mts` (multivariate time series) object, the aggregation matrix and the size of the hierarchy. 
2. Take the `mts` component of the list, which contains the time series data.
3. The first column contains the aggregated data for all cantons; we will drop it for now and we will reconstruct it later.
4. Convert the remaining time series into a tsibble using `as_tsibble()`.
5. Optionally, rename the columns for clarity with `rename()`.
6. Aggregate the data by summing the number of tourist in each canton, using `aggregate_key()`.


```{r data}
data <- swiss_tourism$ts[, -1, drop = FALSE] |>             # Remove 1st column
  as_tsibble() |>                                           # Convert in tsibble
  rename(Month = index, Canton = key, Tourists = value) |>  # Rename
  aggregate_key(Canton, Tourists = sum(Tourists))           # Aggregate all cantons

data |>  head(5) |> kable()
```

The resulting object `data`, contains time series, stacked vertically, with the time index (column `Month`),  the grouping variable (column `Canton`), and the time series values (column `Tourists`). For aggregated time series, the column `Canton` is set to `<aggregated>`.

Note that `fabletools` allows more complex hierarchies with more grouping variables, however here we reproduce the hierarchy in @carrara2025.



## Models fitting

We remove from each time series the last 12 observations, which will be used as test set. We use the function `filter()`  and the `yearmonth()` format to specify the cutoff date. This allows us to evaluate the performance of the reconciliation method on the holdout set.

Then, we fit a base forecasting model to the data with an ETS model with additive trend and seasonality; this is implemented in `fable` with function `ETS()`. We will reconcile the results by using the methods `fable.bayesRecon::bayesRecon_t` (t-Rec) and `fabletools::min_trace` (MinT @wickramasuriya2019optimal). 

```{r fit}
fit <- data |>
  filter(Month < yearmonth("2024 Feb")) |>                  # Remove test set
  model(base = ETS(Tourists ~ trend("A") + season("A"))) |> # fit ETS model
  reconcile(t = bayesRecon_t(base),                         # Reconcile with t-Rec
            mint = min_trace(base))                         # Reconcile with MinT

fit |> head(5) |> kable()
```


During this step, base forecasts are fitted, but predictions are not generated yet.
Thus, we specificy the reconciliation strategy to set up the model, however, the actual reconciliation will be performed when we produce forecasts.

## Reconciliation

Given the fitted models, we can now make one-year ahead predictions.
This is done applying the `forecast()` method to the fitted model.
The forecast horizon can either be specified as a number of periods (e.g., `h = 12` for 12 months) or as a date (e.g., `h = "1 year"`).

```{r forecast}
fc <- fit |>
  forecast(h = "1 year")

# Showing only the one step ahead forecast
fc |> filter(Month == yearmonth("Feb 2024")) |> head(5) |> kable()
```

The forecasts object contains both base (`base`) and reconciled forecasts (`mint` or `t`), specified in the `.model` column.
As for the fitted models, `Canton` and `Month` are respectively the name of the time series and the time index.
The column `Tourists` contains the probabilistic forecasts, encoded as objects from the `distributional` package. Note that both `base` and `mint` models return Normal distributions while `t` returns a Student-t distribution. 
Point forecasts are returned as well: the `.mean` column contains the mean of the predictive distribution.

## Summary of results

The following pipe computes the average performance, according to three accuracy measures, grouped by model:

```{r accuracy}
results <- fc |>
  accuracy(data, measures = list(RMSSE = RMSSE, MASE = MASE, CRPS = CRPS)) |>
  group_by(.model) |>
  summarise(RMSSE = mean(RMSSE), MASE  = mean(MASE), CRPS  = mean(CRPS))

results |> kable(digits = 3)
```



## Visualisation

We can visualize the forecasts using the function `autoplot()`. Below we plot the aggregated series showing both base and reconciled forecasts against the actual data. 

```{r plot-aggregated, fig.width=7, fig.height=4}
# Aggregated series
fc |>
  filter(is_aggregated(Canton)) |>
  autoplot(data, level=95) +
  scale_y_continuous(labels = label_number(suffix = "K", scale = 1e-3))+
  labs(title = "Aggregated Swiss Tourism", y = "Tourists", x = NULL) +
  theme_minimal()
```

The plot below shows a zoomed in version of the forecasts for the aggregated series. The prediction intevals (95% confidence) of both MinT and t-Rec are smaller than the base forecasts. This is because both methods produce coherent forecasts, which are more accurate than the base forecasts for the aggregated series, as shown in the accuracy table above.

```{r plot-aggregated-zoom, fig.width=7, fig.height=4}
# Zoom on the forecasts for the aggregated series
fc |>
  filter(is_aggregated(Canton)) |>
  autoplot(data |> filter(Month >= yearmonth("2022 Jan")), level = 95) +
  scale_y_continuous(labels = label_number(suffix = "K", scale = 1e-3))+
  labs(title = "Aggregated Swiss Tourism (Jan 2022 - Jan 2025)", y = "Tourists", x = NULL) +
  theme_minimal()
```

We now show the forecasts for two bottom level time series, i.e. two cantons: Ticino and Luzern. Both plots are restricted to the last 3 years. 


```{r plot-canton1, fig.width=7, fig.height=4}
# Canton Ticino
fc |>
  filter(Canton == "Ticino") |>
  autoplot(data |> filter(Month >= yearmonth("2022 Jan")),
           level=95) +
  scale_y_continuous(labels = label_number(suffix = "K", scale = 1e-3))+
  labs(title = paste("Canton: Ticino"), y = "Tourists", x = NULL) +
  theme_minimal()
```



```{r plot-canton2, fig.width=7, fig.height=4}
# Canton Luzern
fc |>
  filter(Canton == "Luzern") |>
  autoplot(data |> filter(Month >= yearmonth("2022 Jan"),
                                 Canton == "Luzern"),
           level = 80) +
  scale_y_continuous(labels = label_number(suffix = "K", scale = 1e-3))+
  labs(title = paste("Canton: Luzern"), y = "Tourists", x = NULL) +
  theme_minimal()
```


# References
<div id="refs"></div>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayesRecon_BUIS.R
\name{forecast.lst_bayesRecon_BUIS}
\alias{forecast.lst_bayesRecon_BUIS}
\title{forecast.lst_bayesRecon_BUIS}
\usage{
\method{forecast}{lst_bayesRecon_BUIS}(
  object,
  key_data,
  point_forecast = list(.mean = mean),
  new_data = NULL,
  ...
)
}
\arguments{
\item{object}{An object of class \code{lst_bayesRecon_BUIS} containing fitted models.}

\item{key_data}{A keyed data frame from \code{fabletools}.}

\item{point_forecast}{A list of point forecast functions (default: \code{list(.mean = mean)}).}

\item{new_data}{Optional new data for forecasting (not currently used).}

\item{...}{Additional arguments passed to other methods.}
}
\value{
A fable object with BUIS-reconciled distributions and point forecasts.
}
\description{
Produces probabilistic forecasts reconciled via Bottom-Up Importance Sampling (BUIS).
This method samples from bottom-level distributions, computes upper-level paths via
aggregation, and reweights samples according to upper-level forecast densities.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayesRecon_t.R
\name{bayesRecon_t}
\alias{bayesRecon_t}
\title{t-Rec: reconciliation via conditioning with uncertain covariance via multivariate t-distribution}
\usage{
bayesRecon_t(models, ...)
}
\arguments{
\item{models}{A list of fitted models to reconcile.}

\item{...}{Additional arguments passed to other methods, including:
\itemize{
\item \code{prior}: Optional list with entries \code{nu} and \code{Psi} specifying the parameters
of the Inverse-Wishart prior distribution for the covariance matrix.
If not provided, the prior is estimated from the data.
\item \code{freq}: Optional frequency of the time series, used for estimating the naive covariance matrix
via seasonal naive residuals. If not provided, the frequency is inferred from the data.
\item \code{criterion}: Criterion for estimating the naive covariance matrix (default: "RSS").
\item \code{l_shr}: Optional shrinkage parameter (between 0 and 1) for the covariance matrix of the residuals.
}}
}
\value{
A model specification of class \code{"bayesRecon_t"}
(inheriting from \code{"mdl_lst"}) to be passed to
\code{\link[fabletools]{reconcile}}. The reconciliation is
performed when \code{\link[fabletools]{forecast}} is called on
the resulting mable (a model table). The forecast output is a \code{fable} object
containing reconciled probabilistic forecasts represented as
sample distributions (\code{\link[distributional]{dist_student_t}}).
}
\description{
Specifies t-Rec reconciliation for use within \code{reconcile()}.

Reconciles base forecasts by conditioning on the hierarchical constraints.
The base forecasts are assumed to be jointly Gaussian,
conditionally on the covariance matrix of the forecast errors.
A Bayesian approach is adopted to account for the uncertainty of the covariance matrix.
An Inverse-Wishart prior is specified on the covariance matrix,
leading to a multivariate t-distribution for the base forecasts.
The reconciliation via conditioning is in closed-form,
yielding a multivariate t reconciled distribution.

Reconciliation is performed when \code{forecast()} is called on the resulting model.
Marginal reconciled forecasts follow a Student-t distribution.
}
\examples{
library(fable)
library(fabletools)
library(tsibble)
library(fable.bayesRecon)

# Small hierarchical example using tourism data
tourism_small <- tsibble::tourism |>
dplyr::filter(Region == "Melbourne") |> 
fabletools::aggregate_key(Purpose, Trips = sum(Trips))

fit <- tourism_small |> 
model(base = ETS(Trips ~ trend("A") + season("A"))) |>
reconcile(recon = bayesRecon_t(base))

fc <- forecast(fit, h = "2 years")
fc


}
\references{
Carrara, C., Corani, G., Azzimonti, D., & Zambon, L. (2025). Modeling the uncertainty on the covariance
matrix for probabilistic forecast reconciliation. arXiv preprint arXiv:2506.19554.
\url{https://arxiv.org/abs/2506.19554}
}
\seealso{
\code{\link[fabletools:reconcile]{fabletools::reconcile()}}, \code{\link[fabletools:aggregate_key]{fabletools::aggregate_key()}},
\code{\link[fabletools:min_trace]{fabletools::min_trace()}}, \code{\link[bayesRecon:reconc_t]{bayesRecon::reconc_t()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayesRecon_TDcond.R
\name{forecast.lst_bayesRecon_TDcond}
\alias{forecast.lst_bayesRecon_TDcond}
\title{forecast.lst_bayesRecon_TDcond}
\usage{
\method{forecast}{lst_bayesRecon_TDcond}(
  object,
  key_data,
  point_forecast = list(.mean = mean),
  new_data = NULL,
  ...
)
}
\arguments{
\item{object}{An object of class \code{lst_bayesRecon_MixCond} containing fitted models.}

\item{key_data}{A keyed data frame from \code{fabletools}.}

\item{point_forecast}{A list of point forecast functions (default: \code{list(.mean = mean)}).}

\item{new_data}{Optional new data for forecasting (not currently used).}

\item{...}{extra parameters to be passed on.}
}
\value{
A fable object with TDcond-reconciled distributions and point forecasts.
}
\description{
Produces probabilistic forecasts reconciled via top-down conditioning.
Upper base forecasts are reconciled via conditioning. The recondiled
upper forecasts are then propagate to bottom via probabilisitc top-down.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayesRecon_TDcond.R
\name{make_PMF}
\alias{make_PMF}
\title{Create a discrete PMF from distributions}
\usage{
make_PMF(dist, negative_to_zero = FALSE, toll = 1e-09, alpha_smoothing = 1e-09)
}
\arguments{
\item{dist}{A vector or list of \code{distributional} objects.}

\item{negative_to_zero}{Logical, currently unused. Reserved for future use.}

\item{toll}{Tail probability used to truncate the support when estimating the PMF.}

\item{alpha_smoothing}{Numeric, currently unused. Reserved for future use.}
}
\value{
A list of numeric vectors, one PMF per input distribution.
}
\description{
Converts a set of distributional objects into a list of discrete PMFs used by
the top-down conditioning reconciliation. The function enforces nonnegative,
integer support by truncating negative supports and rounding non-integer
supports via sampling, then estimates the PMF up to the (1 - toll)
quantile.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayesRecon_MixCond.R
\name{forecast.lst_bayesRecon_MixCond}
\alias{forecast.lst_bayesRecon_MixCond}
\title{forecast.lst_bayesRecon_MixCond}
\usage{
\method{forecast}{lst_bayesRecon_MixCond}(
  object,
  key_data,
  point_forecast = list(.mean = mean),
  new_data = NULL,
  ...
)
}
\arguments{
\item{object}{An object of class \code{lst_bayesRecon_MixCond} containing fitted models.}

\item{key_data}{A keyed data frame from \code{fabletools}.}

\item{point_forecast}{A list of point forecast functions (default: \code{list(.mean = mean)}).}

\item{new_data}{Optional new data for forecasting (not currently used).}

\item{...}{Additional arguments passed to other methods.}
}
\value{
A fable object with MixCond-reconciled distributions and point forecasts.
}
\description{
Produces probabilistic forecasts reconciled via mixed hierarchy conditioning.
This method handles mixed hierarchies (with cross-constraints) by applying
conditioning-based reconciliation to draw samples from the reconciled distributions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayesRecon_MixCond.R, R/bayesRecon_TDcond.R
\name{bayesRecon_MixCond}
\alias{bayesRecon_MixCond}
\alias{bayesRecon_TDcond}
\title{Probabilistic reconciliation of mixed hierarchies}
\usage{
bayesRecon_MixCond(models, n_samples = 1000, suppress_warnings = TRUE)

bayesRecon_TDcond(models, n_samples = 1000, suppress_warnings = TRUE)
}
\arguments{
\item{models}{A list of fitted models to reconcile.}

\item{n_samples}{Number of samples to draw from the reconciled distribution.}

\item{suppress_warnings}{If \code{TRUE}, suppress warnings from reconciliation.}
}
\value{
A model specification of class \code{"bayesRecon_MixCond"} or
\code{"bayesRecon_TDcond"} (inheriting from \code{"mdl_lst"})
to be passed to \code{\link[fabletools]{reconcile}}. The reconciliation is
performed when \code{\link[fabletools]{forecast}} is called on
the resulting mable (a model table). The forecast output is a \code{fable} object
containing reconciled probabilistic forecasts represented as
sample distributions (\code{\link[distributional]{dist_sample}}).
}
\description{
\code{bayesRecon_MixCond} specifies Mixed-Conditioning (Mix-Cond) reconciliation for use within
\code{reconcile()}. The method uses importance sampling to draw samples from the reconciled
forecast distribution, obtained via conditioning, in the case of a mixed hierarchy.

\code{bayesRecon_TDcond} specifies Top-Down Conditioning reconciliation for use within
\code{reconcile()}. The method uses a top-down conditioning algorithm: first, upper base forecasts
are reconciled via conditioning using only the hierarchical constraints between the
upper; then, the bottom distributions are updated via a probabilistic top-down procedure.

Reconciliation is performed when \code{forecast()} is called on the resulting model.
Marginal reconciled forecasts follow a sample distribution.
}
\examples{
library(tsibble)
library(dplyr)
library(tibble)
library(fable)
library(fabletools)


# Mixed hierarchy with integer-valued bottom series and one upper aggregate.
# Two low-rate Poisson count series at the bottom; their sum forms the
# continuous-looking upper aggregate. 
set.seed(42)
n <- 60
idx <- tsibble::yearmonth("2019 Jan") + 0:(n - 1)

counts <- dplyr::bind_rows(
   tibble::tibble(Month = idx, Item = "A", Sales = rpois(n, lambda = 3)),
   tibble::tibble(Month = idx, Item = "B", Sales = rpois(n, lambda = 5))) |>
   tsibble::as_tsibble(index = Month, key = Item) |>
   fabletools::aggregate_key(Item, Sales = sum(Sales))
   
# Fit a base model on every level, then reconcile via MixCond:
# the upper (aggregated) forecast is treated as Gaussian, while the
# bottom-level forecast samples are treated as discrete via importance
# sampling.
fit <- counts |>
model(base = ETS(Sales)) |>
reconcile(mc = bayesRecon_MixCond(base))


# Alternative reconciliation via TDcond
fit_TDcond <- counts |>
model(base = ETS(Sales)) |>
reconcile(mc = bayesRecon_TDcond(base))


fc <- forecast(fit, h = 2)
fc


}
\references{
Zambon, L., Azzimonti, D., Rubattu, N., Corani, G. (2024).
\emph{Probabilistic reconciliation of mixed-type hierarchical time series}.
Proceedings of the Fortieth Conference on Uncertainty in Artificial Intelligence,
PMLR 244:4078-4095. \url{https://proceedings.mlr.press/v244/zambon24a.html}.
}
\seealso{
\code{\link[fabletools:reconcile]{fabletools::reconcile()}}, \code{\link[fabletools:aggregate_key]{fabletools::aggregate_key()}},
\code{\link[=bayesRecon_BUIS]{bayesRecon_BUIS()}}, \code{\link[bayesRecon:reconc_MixCond]{bayesRecon::reconc_MixCond()}}, \code{\link[bayesRecon:reconc_TDcond]{bayesRecon::reconc_TDcond()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bayesRecon_BUIS.R
\name{bayesRecon_BUIS}
\alias{bayesRecon_BUIS}
\title{BUIS for probabilistic reconciliation of forecasts via conditioning}
\usage{
bayesRecon_BUIS(models, n_samples = 1000)
}
\arguments{
\item{models}{A list of fitted models to reconcile.}

\item{n_samples}{Number of samples to draw from the reconciled distribution.}
}
\value{
A model specification of class \code{"bayesRecon_BUIS"}
(inheriting from \code{"mdl_lst"}) to be passed to
\code{\link[fabletools]{reconcile}}. The reconciliation is
performed when \code{\link[fabletools]{forecast}} is called on
the resulting mable (a model table). The forecast output is a \code{fable} object
containing reconciled probabilistic forecasts represented as
sample distributions (\code{\link[distributional]{dist_sample}}).
}
\description{
Specifies Bottom-Up Importance Sampling (BUIS) reconciliation for use within
\code{reconcile()}. The method uses the Bottom-Up Importance Sampling algorithm
to draw samples from the reconciled forecast distribution, obtained via conditioning.
Reconciliation is performed when \code{forecast()} is called on the resulting model.
Marginal reconciled forecasts follow a sample distribution.
}
\examples{
library(fable)
library(fabletools)
library(tsibble)
library(fable.bayesRecon)

# Small hierarchical example using tourism data
tourism_small <- tsibble::tourism |>
dplyr::filter(Region == "Melbourne") |> 
fabletools::aggregate_key(Purpose, Trips = sum(Trips))

fit <- tourism_small |> 
model(base = ETS(Trips ~ trend("A") + season("A"))) |>
reconcile(recon = bayesRecon_BUIS(base, n_samples = 300))

fc <- forecast(fit, h = "2 years")
fc

}
\references{
Zambon, L., Azzimonti, D. & Corani, G. (2024).
\emph{Efficient probabilistic reconciliation of forecasts for real-valued and count time series}.
Statistics and Computing 34 (1), 21.
\doi{10.1007/s11222-023-10343-y}.
}
\seealso{
\code{\link[fabletools:reconcile]{fabletools::reconcile()}}, \code{\link[fabletools:aggregate_key]{fabletools::aggregate_key()}}, \code{\link[=bayesRecon_MixCond]{bayesRecon_MixCond()}}, \code{\link[bayesRecon:reconc_BUIS]{bayesRecon::reconc_BUIS()}}
}
